from PyFaceGen.Mesh.Mesh import Mesh
from PyFaceGen.Mesh.Utils import get_normals
import numpy as np

def check_topology_is_shared(a,b) -> bool:
    
    """ Checks if two meshes share the same number of vertices and faces """
    
    if a.vertices.shape != b.vertices.shape:
        return False
    if a.faces.shape != b.faces.shape:
        return False
    return True

def deformation_transfer(source_neutral, source_expression, target_neutral) -> Mesh:
    
    """ Using deformation transfer for triangle meshes (Sumner & Popovic)
    http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.126.6553&rep=rep1&type=pdf
    Apply the expression of a source identity to a target identity
    All meshes must share a common topology
    
    (params):
        source_neutral (Mesh): The source identity in neutral expression
        source_expression (Mesh): The source identity with target expression
        target_neutral (Mesh): The identity of the target with neutral expression
    
    (returns):
        (Mesh): The target identity with the source expression transferred onto it
    """
    
    # Get the number of vertices and faces
    n_faces = source_neutral.faces.shape[0]
    n_verts = source_neutral.vertices.shape[0]
    
    # Ensure that the meshes are in the same topology
    if not (check_topology_is_shared(source_neutral, source_expression) 
            and check_topology_is_shared(source_neutral, target_neutral)):
        raise RuntimeError('Meshes must share a common topology')
    
    print("Computing B Matrix")
    # B matrix
    B = np.zeros((n_faces * 3, 3))
    
    # Create the B matrix iterating over faces
    for f_idx in range(n_faces):
        
        # Get the source neutral vertices in this face
        a,b,c = source_neutral.faces[f_idx, :, 0]
        v1 = source_neutral.vertices[a, :]
        v2 = source_neutral.vertices[b, :]
        v3 = source_neutral.vertices[c, :]
        
        # v4 calculation
        cross = np.cross((v2 - v1), (v3 - v1))  
        v4 = v1 + cross/np.sqrt(np.linalg.norm(cross))
        
        # V matrix
        V = np.zeros((3,3))
        V[:,0] = v2 - v1
        V[:,1] = v3 - v1
        V[:,2] = v4 - v1
        
        # Get the source expression vertices in this face
        v1_ = source_expression.vertices[a, :]
        v2_ = source_expression.vertices[b, :]
        v3_ = source_expression.vertices[c, :]
        
        # v4 bar calculation
        cross_ = np.cross((v2_ - v1_), (v3_ - v1_))  
        v4_ = v1_ + cross/np.sqrt(np.linalg.norm(cross_))
        
        # V bar matrix
        V_ = np.zeros((3,3))
        V_[:,0] = v2_ - v1_
        V_[:,1] = v3_ - v1_
        V_[:,2] = v4_ - v1_
        
        # Q matrix calculation
        Q = V_ @ np.linalg.inv(V)
        QT = Q.T
        
        # Update B
        offset = f_idx * 3
        B[offset: offset+3, :] = QT
    
    print("Calculating A Matrix")
    A = np.zeros((n_faces * 3, n_faces + n_verts))
    for f_idx in range(n_faces):
        
        # Get the target neutral vertices in this face
        a,b,c = target_neutral.faces[f_idx, :, 0]
        v1 = target_neutral.vertices[a, :]
        v2 = target_neutral.vertices[b, :]
        v3 = target_neutral.vertices[c, :]
        
        # v4 calculation
        cross = np.cross((v2 - v1), (v3 - v1))  
        v4 = v1 + cross/np.sqrt(np.linalg.norm(cross))  
        
        # V matrix
        V = np.zeros((3,3))
        V[:,0] = v2 - v1
        V[:,1] = v3 - v1
        V[:,2] = v4 - v1
        
        V_inv = np.linalg.inv(V)

        # Update A
        offset = f_idx * 3
        for col in range(3):
            A[offset+col, a] = -(V_inv[0, col] + V_inv[1, col] + V_inv[2, col])
            A[offset+col, b] = V_inv[0, col]
            A[offset+col, c] = V_inv[1, col]
            A[offset+col, n_verts + f_idx] = V_inv[2, col]
            
    print("Setting up linear system")
    AT = A.T
    ATA = AT @ A
    ATB = AT @ B
    
    # Clear memory
    #del(A)
    #del(B)
    del(AT)
    
    print("Solving linear system")
    
    X = np.linalg.solve(ATA, ATB)
    # Clear memory
    del(ATA)
    del(ATB)
    
    new_verts = np.zeros_like(target_neutral.vertices)
    new_uvs = np.array(target_neutral.UVs)
    new_faces = np.array(target_neutral.faces)
    
    # TODO: Update these via X
    new_normals = get_normals(new_verts, new_faces)
    
    
    for v_idx in range(n_verts):
        new_verts[v_idx, :] = X[v_idx, :]
        
    return Mesh(new_verts, new_uvs, new_normals, new_faces)
        
    
    
        
        


#from PyFaceGen.IO.ply_loader import ply_2_mesh, mesh_2_ply
from PyFaceGen.IO.wobj_loader import wobj_2_mesh, mesh_2_wobj
import numpy as np
# ---------------------- Ply Tests ------------------------- #

'''# Test case 1 - can we load a simple obj
def plytestsimple():
    mesh = ply_2_mesh('Data/Tests/ply.ply')
    assert mesh.vertices[0][0] != 0


# Test case 2 - can we load a ply file with binary encoding    
def plytestbinary():
    
    mesh = ply_2_mesh('Data/Tests/ply_binary.ply')
    assert mesh.vertices[0][0] != 0
    
# Run all ply tests    
def plytests():
    
    plytestsimple()
    plytestbinary()'''

# Test case 1 - can we load a simple wavefront obj file
def wobjtestsimple():
    mesh = wobj_2_mesh('Data/Tests/wobj.obj')
    assert mesh.vertices[0][0] != 0

# Test case 2 - does reading/writing UVs work    
def wobjtestUVs():
    mesh = wobj_2_mesh('Data/Tests/wobj.obj')
    assert mesh.UVs[0][0] != 0
    mesh.UVs[0][0] = 0.5
    mesh_2_wobj(mesh, 'Data/Tests/wobj_uv.obj')
    mesh = wobj_2_mesh('Data/Tests/wobj_uv.obj')
    assert  mesh.UVs[0][0] == 0.5
    
# Test case 2 - does reading/writing normals work    
def wobjtestnormals():
    mesh = wobj_2_mesh('Data/Tests/wobj.obj')
    print(mesh.faces.shape)
    assert mesh.normals[0][0] != 0
    mesh.normals[0][0] = 0.5
    mesh_2_wobj(mesh, 'Data/Tests/wobj_norm.obj')
    mesh = wobj_2_mesh('Data/Tests/wobj_norm.obj')
    assert  mesh.normals[0][0] == 0.5
    

# Run all wobj tests
def wobjtests():
    wobjtestsimple()
    wobjtestUVs()
    wobjtestnormals()




from PyFaceGen.Mesh.Mesh import Mesh
from PyFaceGen.Mesh.Graph import MeshGraph

import numpy as np

def get_components_given_vertex(mesh, vert):
    
    graph = MeshGraph(mesh)
    
    component = set([])
    frontier = set([vert])
    while len(frontier) != 0:
        x = frontier.pop()
        component.add(x)
        for sink in graph.edges[x]:
            frontier.add(sink)
        print(len(frontier))    
    return component

        
def get_components(mesh):
    
    # The idea is that we will create components of verex indicies
    # and merge them if there is a face that contains verticies from different components
    
    components = []
    faces = mesh.faces[:,:,0]
    
    for i, face in enumerate(faces):
        a,b,c = face
        a_comp, b_comp, c_comp = -1,-1,-1 # -1 means no exiting component contains them
        
        # Find which, if any, components each a,b and c already belong to
        for j,component in enumerate(components):
            
            # Iterate over each vertex
            for vertex in component:
                
                if vertex == a:
                    a_comp = j
                if vertex == b:
                    b_comp = j
                if vertex == c:
                    c_comp = j
                    
        # Now we consider a few cases
        
        # Firstly if none of the vertices are already assigned to a component
        if a_comp == b_comp == c_comp == -1:
            # In this case, make a new one for them
            components.append([a,b,c])
            
        # Next if only one vertex belongs to a component then assign all other vertices to it
        elif a_comp != -1 and b_comp == c_comp == -1:
            components[a_comp].append(b)
            components[a_comp].append(c)
            
        elif b_comp != -1 and a_comp == c_comp == -1:
            components[b_comp].append(a)
            components[b_comp].append(c)
            
        elif c_comp != -1 and b_comp == a_comp == -1:
            components[c_comp].append(b)
            components[c_comp].append(a)
            
        # Next if any two vertices already have a component 
        # If its the same we add the third vertex to it
        # If not we merge the components and add the other vertex
        elif a_comp != -1 and b_comp != -1 and c_comp == -1:
            # Merge and delete
            if a_comp != b_comp:
                # Add all of component b to component a along with c and remove component b
                components[a_comp].extend(components[b_comp])
                components[a_comp].append(c)
                components.pop(b_comp)
            # Just add the remaining vertex
            else:
                components[a_comp].append(c)
                
        elif b_comp != -1 and c_comp != -1 and a_comp == -1:
            # Merge and delete
            if b_comp != c_comp:
                # Add all of component c to component b along with a and remove component c
                components[b_comp].extend(components[c_comp])
                components[b_comp].append(a)
                components.pop(c_comp)
            # Just add the remaining vertex
            else:
                components[b_comp].append(a)
        elif c_comp != -1 and a_comp != -1 and b_comp == -1:
            # Merge and delete
            if c_comp != a_comp:
                # Add all of component a to component c along with b and remove component a
                components[c_comp].extend(components[a_comp])
                components[c_comp].append(b)
                components.pop(a_comp)
            # Just add the remaining component
            else:
                components[c_comp].append(b)
                
        # Finally if all vertices already belong to components
        # Merge any components that are seperate
        elif a_comp != -1 and b_comp != -1 and c_comp != -1:
            
            if a_comp == b_comp == c_comp:
                continue #do nothing, all already fine
            
            # Merge component c into the component containing both a and b 
            if a_comp == b_comp != c_comp:
                components[a_comp].extend(components[c_comp])
                components.pop(c_comp)
                
            # Merge component a into the component containing both b and c 
            elif a_comp != b_comp == c_comp:
                components[c_comp].extend(components[a_comp])
                components.pop(a_comp)
                
            # Merge component b into the component containing both a and c 
            elif a_comp == c_comp != b_comp:
                components[a_comp].extend(components[b_comp])
                components.pop(b_comp)
            
            # Everything is different, we have to be careful removing 2 components
            elif a_comp != b_comp != c_comp:
                components[a_comp].extend(components[b_comp])
                components[a_comp].extend(components[c_comp])
                
                copy = [] # We need to do this as a copy as removing would change indicies
                for c_id, component in enumerate(components):
                    if c_id not in [b_comp, c_comp]:
                        copy.append(component)
                components = copy
                
            else:
                # We shouldn't get here if this code is correct
                raise RuntimeError('Components %s, %s, %s should not be valid'%(a_comp, b_comp, c_comp))
        else:
            # We shouldn't get here if this code is correct
            raise RuntimeError('Components %s, %s, %s should not be valid'%(a_comp, b_comp, c_comp))

    return components

def split_by_components(mesh, components):
    
    # Convert vertices
    vertices = [mesh.vertices[component] for component in components]
    normals = [[] for component in components]
    uvs = [[] for component in components]
    faces = [[] for component in components]
    uv_ids = [[] for component in components]
    n_ids = [[] for component in components]
    
    for face in mesh.faces:
        a,b,c = face[:,0] # The split is based on position
        ta,tb,tc = face[:,1] # We still want the UV though
        na,nb,nc = face[:,2] # And normals
        
        for i,component in enumerate(components):
            
            # Find the component assositated with this face
            id_a, id_b, id_c = -1,-1,-1  # These will be -1 if this is the wrong componemt
            
            for j,v in enumerate(component):
                if v==a:
                    # We have found the right component
                    id_a = j # j will be the index of the vertex in the split vertex list
                elif v==b:
                    id_b = j
                elif v==c:
                    id_c = j
            if id_a == id_b == id_c == -1:
                # Wrong component 
                continue
            else:
                
                # We have found the component this face is in
                if -1 in [id_a, id_b, id_c]:
                    # This is a failsafe, shouldn't happen
                    raise RuntimeError('This shouldnt happen, components has failed')
                    
                uv_a, uv_b, uv_c = face[:,1]
                n_a, n_b, n_c = face[:,2]
                
                # Ensure we dont duplicate the uvs
                if uv_a not in uv_ids[i]:
                    uv_ids[i].append(uv_a)
                    uvs[i].append(mesh.UVs[uv_a])
                    uv_id_a = len(uvs[i]) - 1
                else:
                    uv_id_a = uv_ids[i].index(uv_a)
                    
                if uv_b not in uv_ids[i]:
                    uv_ids[i].append(uv_b)
                    uvs[i].append(mesh.UVs[uv_b])
                    uv_id_b = len(uvs[i]) - 1
                else:
                    uv_id_b = uv_ids[i].index(uv_b)
                    
                if uv_c not in uv_ids[i]:
                    uv_ids[i].append(uv_c)
                    uvs[i].append(mesh.UVs[uv_c])
                    uv_id_c = len(uvs[i]) - 1
                else:
                    uv_id_c = uv_ids[i].index(uv_c)
                    
                
                # Ensure we dont duplicate the normals
                if n_a not in n_ids[i]:
                    n_ids[i].append(n_a)
                    normals[i].append(mesh.normals[n_a])
                    n_id_a = len(normals[i]) - 1
                else:
                    n_id_a = n_ids[i].index(n_a)
                    
                if n_b not in n_ids[i]:
                    n_ids[i].append(n_b)
                    normals[i].append(mesh.normals[n_b])
                    n_id_b = len(normals[i]) - 1
                else:
                    n_id_b = n_ids[i].index(n_b)
                    
                if n_c not in n_ids[i]:
                    n_ids[i].append(n_c)
                    normals[i].append(mesh.normals[n_c])
                    n_id_c = len(normals[i]) - 1
                else:
                    n_id_c = n_ids[i].index(n_c)
                    
                
                    
                
                faces[i].append(np.array([[id_a,id_b,id_c],
                                          [uv_id_a, uv_id_b, uv_id_c],
                                          [n_id_a, n_id_b, n_id_c]]).T)

                
    return [Mesh(np.array(vertices[i]), np.array(uvs[i]), np.array(normals[i]), np.array(faces[i])) for i in range(len(components))]



from PyFaceGen.Mesh.Mesh import Mesh
from PyFaceGen.Mesh.Utils import get_normals
import numpy as np

# Wavefront loader
def wobj_2_mesh(path):
    
    # Initialise the properties
    vertices = []
    normals = []
    uvs = []
    faces = []
    n_faces = 0
    with open(path,'r') as f:
        
        for line in f:
            
            elements = line.split(' ')
            
            if elements[0] == 'v':
                # Vertex position
                x,y,z = float(elements[1]), float(elements[2]), float(elements[3])
                vertex = np.array([x,y,z])
                vertices.append(vertex)
                
            elif elements[0] == 'vt':
                # UV coordinate
                u,v = float(elements[1]), float(elements[2])
                coord = np.array([u,v])
                uvs.append(coord)
                
            elif elements[0] == 'vn':
                # Normal vector
                x,y,z = float(elements[1]), float(elements[2]), float(elements[3])
                normal = np.array([x,y,z])
                normals.append(normal)
                
            elif elements[0] == 'f':
                # Face
                n_faces += 1
                
                if len(elements) > 4:
                    # Catch trailing spaces
                    if ''in elements:
                        elements.remove('')
                
                # Get the indcies of the face
                a,b,c = elements[1], elements[2], elements[3]
                
                if len(a.split('/')) == 1:
                    # No UV or normal information
                    faces.append(np.array([[int(a)-1,int(b)-1,int(c)-1]]).T)
                    
                elif len(a.split('/')) == 2:
                    # We have uv info
                    a_vert, a_uv = a.split('/')
                    b_vert, b_uv = b.split('/')
                    c_vert, c_uv= c.split('/')
                    
                    faces.append(np.array([[int(a_vert)-1,int(b_vert)-1, int(c_vert)-1],
                                           [int(a_uv)-1,int(b_uv)-1, int(c_uv)-1]]).T)
                    
                    
                else:
                    a_vert, a_uv, a_norm = a.split('/')
                    b_vert, b_uv, b_norm = b.split('/')
                    c_vert, c_uv, c_norm = c.split('/')
                    
                    faces.append(np.array([[int(a_vert)-1,int(b_vert)-1, int(c_vert)-1],
                                           [int(a_uv)-1,int(b_uv)-1, int(c_uv)-1],
                                           [int(a_norm)-1,int(b_norm)-1, int(c_norm)-1]]).T)
                    
    # Set the properties to arrays, for simplicity we will always enforce having normals and uvs, even as placeholders          
    vertices = np.array(vertices)
    faces = np.array(faces)
    if normals == []:
        # Compute normals using 1-hop neighbours
        normals = get_normals(vertices, faces)
        f_normal = np.array(faces[:,:,0]).reshape((-1,3,1)) # Vertex postiton faces with normals in 1-1 correspondance
        faces = np.concatenate((faces, f_normal), axis=2)
      
    # We always now have at least position and normals for faces    
    else:    
        normals = np.array(normals)
        
    if uvs == []:
        # Set all UV coords to 0,0
        uvs = np.zeros((vertices.shape[0], 2))
        f_uvs = np.zeros((n_faces, 3, 3), dtype='uint32')
        f_uvs[:,:,0] = faces[:,:,0]
        f_uvs[:,:,1] = faces[:,:,0]
        f_uvs[:,:,2] = faces[:,:,1]
        faces = f_uvs
        
    else:
        uvs = np.array(uvs)
    
    
    mesh = Mesh(vertices, uvs, normals, faces)
    return mesh
                

def mesh_2_wobj(mesh,path):
    
    ''' Converts a .obj file to the custom mesh class '''
    
    # Get the properties of the mesh
    vertices = mesh.vertices
    uvs = mesh.UVs
    normals = mesh.normals
    faces = mesh.faces
    
    if faces.shape[-1] == 1:
        has_uv = has_norm = False
    elif faces.shape[-1] == 2:
        # Either vertices or normals
        has_uv = True
        has_norm = False
        
    elif faces.shape[-1] == 3:
        has_uv = True
        has_norm = True
        
    
    with open(path, 'w') as f:
        
        # Open a file, write some header information
        
        f.write('# Generated by script \n')
        f.write('# %s vertices, %s faces \n'%(vertices.shape[0], faces.shape[0]))
        
        for vertex in vertices:
            
            f.write('v %.4f %.4f %.4f\n'%(vertex[0], vertex[1], vertex[2]))
        
        if has_uv:
            for uv in uvs:
            
                f.write('vt %.4f %.4f\n'%(uv[0], uv[1]))
            
        if has_norm:
            for normal in normals:
            
                f.write('vn %.4f %.4f %.4f\n'%(normal[0], normal[1], normal[2]))
        
        for face in faces:
            
            # Face is 3 by n_properties
            
            if not has_uv and not has_norm:
            
                f.write('f %i %i %i\n'%(face[0,0]+1, face[1,0]+1, face[2,0]+1))
                
            elif has_uv and not has_norm:
                
                f.write('f %i/%i %i/%i %i/%i \n'%(face[0,0]+1,face[0,1]+1, face[1,0]+1,face[1,1]+1, face[2,0]+1,face[2,1]+1))
            
            elif has_uv and has_norm:
                
                f.write('f %i/%i/%i %i/%i/%i %i/%i/%i \n'%(face[0,0]+1,face[0,1]+1,face[0,2]+1, face[1,0]+1,face[1,1]+1,face[1,2]+1, face[2,0]+1,face[2,1]+1, face[2,2]+1))
                
            else:
                
                # Catch 
                f.write('f %i %i %i\n'%(face[0]+1, face[1]+1, face[2]+1))



# Non-rigid iterative closest point on meshes in 1-1 vertex correspondance
# For best results, apply rigid ICP first then call this code

# Algorithm implimentation of https://www.researchgate.net/publication/200172513_Optimal_Step_Nonrigid_ICP_Algorithms_for_Surface_Registration
# Sparse idea from 
# Schedules for training from https://github.com/shubhamag/non_rigid_icp/blob/master/nicp_meshes.py
# As well as weighting points by the normal relative to the K nearest neighbours

import numpy as np 

from PyFaceGen.Mesh.Mesh import Mesh
from PyFaceGen.Mesh.Graph import MeshGraph
from PyFaceGen.Mesh.Utils import get_normals
from PyFaceGen.IO.wobj_loader import mesh_2_wobj, wobj_2_mesh
from sklearn.neighbors import NearestNeighbors
from scipy import sparse
from sksparse.cholmod import cholesky_AAt

ALPHA_SCHEDULE = [50,35,20,15, 7,3,2]#,1]#, 0.5]#, 0.1]#, 0.05]
BETA_SCHEDULE = [15,14, 3, 2, 0.5, 0,0]#,0]#,0]#,0]#, 0.05]
CHANGE_THRESHOLD = 0.01
NORMAL_THRESHOLD = 1
USE_NORMALS = False
K = 30
FIX_NORMALS = False
DISCARD = 10 # Discard all matches that are worse than this

def get_d_loss(tr_verts, fi_verts, weights=None):
    
    """ Gets the E_d loss between transfomed vertices and the corresponding fixed vertices"""
    
    assert tr_verts.shape[0] == fi_verts.shape[0]
    # Handle uniform weighting
    if weights is None:
        weights = np.ones(tr_verts.shape[0])
    
    # Homogenous -> world if present
    tr_verts = tr_verts[:, 0:3]
    fi_verts = fi_verts[:, 0:3]
    
    norms = np.linalg.norm(tr_verts-fi_verts, axis=1)
    norms *= weights
    
    return np.sum(norms)

def get_s_loss(graph, transform, gamma):
    
    """ Gets the E_s (stiffness) loss """
    G = np.diag([1.0,1.0,1.0,gamma])
    count = 0
    for i in graph.edges:
        for j in graph.edges[i]:
            count += np.linalg.norm((transform[4*i: 4*i + 4, :]-transform[4*j: 4*j + 4, :]).T @ G)
    return count

def get_correspondance(fl_verts, fi_verts):    
    
    """
    Gets the corresponding vertices of the fixed mesh given 
    the transformed target mesh 
    
    """    
    nn = NearestNeighbors(n_neighbors=1).fit(fi_verts)
    dist, idx = nn.kneighbors(fl_verts)
    return fi_verts[idx].reshape(-1,4), idx
    

def construct_MG(faces, n_vertices, gamma):
    
    """ Builds the knoeneker product MG of node arc incidence matrix for the given graph
    with the diagonal matrix [1,1,1,gamma]
        (returns)
            node-arc incidence matrix
            n_edges (int): The number of unique ordered edges 
            
    """
    edges = []
    
    # TODO: Test this
    for face in faces[:,:,0]: # Workimg on positions
        # Every edge should be from lowest index to highest
        sort = np.sort(face)
        edges.append(tuple([sort[0], sort[1]]))
        edges.append(tuple([sort[0], sort[2]]))
        edges.append(tuple([sort[1], sort[2]]))
        
        
    # Remove duplicates
    edges = set(edges)
    n_edges = len(edges)
    print("Number of edges = %s"%(n_edges))
    
    # Template of the matrix
    M = sparse.lil_matrix((n_edges, n_vertices), dtype='float32')
    #M = np.zeros((n_edges, n_vertices), dtype='float32')
    
    # -1 for source +1 for sink
    for idx, edge in enumerate(edges):
        i,j = edge
        M[idx, i] = -1
        M[idx, j] = 1
        
    G = np.diag([1, 1, 1, gamma]).astype(np.float32)
        
    return sparse.kron(M,G), n_edges

def NRICP(floating, fixed, fl_landmarks=None, fi_landmarks=None, verbose=False, mask=None):
    
    """ Perform optimal step non-rigid ICP to register floating mesh to fixed 
        (Amberg & Romdhani)
        https://www.researchgate.net/publication/200172513_Optimal_Step_Nonrigid_ICP_Algorithms_for_Surface_Registration
    
        (params):
            floating (Mesh): The mesh to be aligned
            fixed (Mesh): The target mesh that will not change
            (fl_landmarks) (list): The floating landmarks of the form
                [(idx, a ,b)] where idx is the index of the triangle in the face
                and a and b barycentric coordinates 
            (fi_landmarks) (list): Same as fl_landmarks but for the fixed mesh
            verbose (bool): If to print debugging information
            mask (list of int): The indices of vertices not to be transformed
    """
    
    # Decide if we are using landmarks
    if (fi_landmarks is not None) and (fl_landmarks is not None):
        USE_LANDMARKS = True
    else:
        USE_LANDMARKS = False
    
    # Get the vertices
    fl_verts = floating.vertices
    fi_verts = fixed.vertices
    
    n_vertices = fl_verts.shape[0]
    n_vertices_target = fi_verts.shape[0]
    
    # Convert to homogenous coordinates
    fl_verts = np.concatenate((fl_verts, np.ones((fl_verts.shape[0],1), dtype='float32' )), axis=1)
    fi_verts = np.concatenate((fi_verts, np.ones((fi_verts.shape[0],1), dtype='float32' )), axis=1)
    
    # Initalise the transforms
    transform = np.zeros((4*n_vertices, 3))
    
    for i in range(n_vertices):
        transform[4*i:(4*i)+3,:] = np.eye(3)
    
    # M kron G
    MG, n_edges = construct_MG(floating.faces, n_vertices, 1)   
    
    # For correspondances
    nn = NearestNeighbors(n_neighbors=1, algorithm='kd_tree').fit(fi_verts[:,0:3])
    knn = NearestNeighbors(n_neighbors=K).fit(fi_verts[:,0:3])
    
    # D matrix
    D = sparse.lil_matrix((n_vertices,n_vertices*4), dtype=np.float32)
    for i in range(n_vertices):
        D[i, 4*i:(4*i)+4] = fl_verts[i,:]
    
    
    # ---------------------------- Landmarks ------------------------------#
    if verbose:
        print("Processing landmarks")
    
    # Prepare the landmarks, they now must map single vertices in the source 
    # to points on the surface of the target
    
    if USE_LANDMARKS:
    
        fi_landmark_points = []
        fl_landmark_indices = []
        for i in range(len(fi_landmarks)):
            fl_land, fi_land = fl_landmarks[i], fi_landmarks[i]
            
            fl_id, fl_a, fl_b = fl_land
            fi_id, fi_a, fi_b = fi_land
            
            # First the floating landmark, we want the closest vertex only
            a,b,c = floating.faces[fl_id,:,0]
            fl_c = 1-(fl_a+fl_b)
            
            # Smallest barycentric corresponds to closest
            if fl_a < fl_c and fl_a < fl_b:
                fl_landmark_indices.append(a)
            
            elif fl_b < fl_a and fl_b < fl_c:
                fl_landmark_indices.append(b)
            
            elif fl_c < fl_a and fl_c < fl_b:
                fl_landmark_indices.append(c)
            
            # Now we do fixed landmarks, here we want the 3D point
            a,b,c = fixed.faces[fi_id,:,0]
            va,vb,vc = fi_verts[a, 0:3], fi_verts[b, 0:3], fi_verts[c, 0:3]
            
            fi_point = (fi_a * va) + (fi_b *vb) +((1- fi_a - fi_b) *  vc)
            fi_landmark_points.append(fi_point)
            
        # Vectorise
        fi_landmark_points = np.array(fi_landmark_points)
        fl_landmark_indices = np.array(fl_landmark_indices)
        
        UL = fi_landmark_points

    # ----------------------- Loop over stiffness --------------------------- #
    
    graph = MeshGraph(floating)
    
    for epoch , alpha in enumerate(ALPHA_SCHEDULE):

        beta = BETA_SCHEDULE[epoch]
        if beta == 0:
            USE_LANDMARKS = False
        
        MAX_ITERS = 5
        if alpha < 25:
            MAX_ITERS = 3
        
        if verbose:
            print("Using stiffness %s"%(alpha))
        
        # Keep track of changes in the transform
        change = np.inf
        n_iter = 0 # Early stop if too many iterations
        
        
        # ------------------ Loop until convergence ------------------------- #
        while change > CHANGE_THRESHOLD:
            
            weights = np.ones((n_vertices,1))
            
            transformed = D @ transform
            
            dist, idx = nn.kneighbors(transformed[:,0:3])
            idx = idx.squeeze()   # https://github.com/saikiran321/nonrigid_icp/blob/master/nricp.py does this and it helps
            correspondance = fi_verts[idx, 0:3]
            
            # Discard any poor matches
            weights[np.where(dist>DISCARD)] = 0 
            
            # If using normals for weighting
            if USE_NORMALS and alpha < 3:
                # Only do it for low stiffness to overcome poor initial registration
                dists = np.zeros(fl_verts.shape[0])
                norms = get_normals(transformed , floating.faces)
                dist_k, idx_k = knn.kneighbors(transformed)
                for i in range(fl_verts.shape[0]):
                    n = norms[i]
                    k_corr = fi_verts[idx_k[i], 0:3] # This gives the KNN of the point
                    diff = fl_verts[i, 0:3] - k_corr # This gets the direction of each KNN relative to the point
                    # We look at the difference in angle between the lines to near points and the normal
                    cosine_difference = np.cross(diff, n) # On a flat surface this should alway be orthogonal
                    
                    filtr = np.linalg.norm(cosine_difference, axis=1) < 0.05 # The larger this is the more orthogonal n is to nearby points
                    if np.sum(filtr) > 0:
                        match = np.mean(k_corr[filtr], axis=0)
                    else:
                        match = 0
                    dists[i] = np.linalg.norm(match-fl_verts[i,0:3])
                        
                weights[np.where(dists>NORMAL_THRESHOLD)[0]] = 0
            
            if mask is not None:
                # Exclude vertices in the mask by setting their weight to 0
                for idx in mask:
                    weights[idx] = 0
                    
            # Weighted match matrix
            U = weights * correspondance
            
            # A,B,U and D Directly from the paper
            A = sparse.vstack([alpha * MG, D.multiply(weights)])
            A = sparse.csr_matrix(A)
            
            B = sparse.lil_matrix((4 * n_edges + n_vertices, 3), dtype=np.float32)
            B[4 * n_edges: (4 * n_edges +n_vertices), :] = U
            B = sparse.csr_matrix(B)
            
            if USE_LANDMARKS:
                # Landmark terms + weighting
                DL = sparse.lil_matrix((UL.shape[0], 4*n_vertices))
                for count, l_idx in enumerate(fl_landmark_indices):
                    DL[count, 4*l_idx: 4*l_idx + 4] = fl_verts[l_idx,:] * beta                    
            
                A = sparse.csr_matrix(sparse.vstack([A, DL]))
                B = sparse.csr_matrix(sparse.vstack([B, beta * UL]))
            
            # This seems to be the only way of solving without lots of RAM    
            factor = cholesky_AAt(A.T)
            X = factor(A.T.dot(B)).toarray()
            
            change = np.linalg.norm(X-transform)
            transform = X
            n_iter += 1
            if n_iter > MAX_ITERS: # Early stop
                break
            
            if verbose:
                print('Change: ' , change)
                print('D loss: ', get_d_loss(D@transform, correspondance))
                print('S loss: ', get_s_loss(graph, transform, 1.0))
                
                if USE_LANDMARKS:
                    print('L loss: ', get_d_loss((D@transform)[fl_landmark_indices, :], UL))
                    print('Loss: ', get_d_loss(D@transform, correspondance) + alpha * get_s_loss(graph, transform, 1.0) + beta * get_d_loss((D@transform)[fl_landmark_indices, :], UL))
                
                else:
                    print('Loss: ', get_d_loss(D@transform, correspondance) + alpha * get_s_loss(graph, transform, 1.0))
    
    # Copy face and UV information from the ols mesh        
    out_faces = floating.faces
    out_uvs = floating.UVs
    out_normals = floating.normals
    
    out_vertices = D @ transform
    
    centroid = np.mean(out_vertices, axis=0)
    for i in range(n_vertices):
        
        # Apply the transform to the normals
        
        n = out_normals[i, :] # 3x1
        x = transform[4*i:(4*i)+3,:] # 3x3 no translation
        n_ = np.linalg.inv(x).T @ n
        n_ /= np.linalg.norm(n_)
        
        # If we are told to, try to flip some incorrect normals
        if FIX_NORMALS:
            
            # TODO: Fix backface issues
            # We'll look at the nearest neighbours and if more
            # than a certain ratio are opposite in direction we flip
            pass
        
        out_normals[i, :] = n_
    return Mesh(out_vertices, out_uvs, out_normals, out_faces), transform
            
            
    



# python version 3.8
# This code is a collection of utility functions

import numpy as np
from math import sin, cos

def cross(a,b):
    
    ''' A 2D cross product '''
    return a[0]*b[1] - a[1]*b[0]


def dot(a,b):
    
    ''' A 2D dot product '''
    return a[0]*b[0] + a[1]*b[1]


class TriangleTester:
    
    def __init__(self, a, b, c):
        
        ''' A class for testing if points belong to the interior of a triangle
            in two dimensions 
            
            (params):
                a, b, c (array-like): The three points in 2D that 
                        describe the triangle
        '''
        
        # By precalculating everything not involving the test point we can
        # speed up the code significantly
        self.ab = cross(a,b)
        self.bc = cross(b,c)
        self.ca = cross(c,a)
        
        self.d = max(1e-14, self.ab + self.bc + self.ca) # To avoid div by 0
        self.scale = 1/self.d
        
        self.amb = a-b
        self.bmc = b-c
        self.cma = c-a
        
        
    def test(self, p):
        
        ''' Tests if the point p is within the interior of the triangle defined
            for this class and gives the barycentric co-ordinates if it is
            (params):
                p (array-like): The point that is to be tested
            (returns):
                w_a, w_b, w_c (float): The barycentric co-ordinates
        '''
        
        if (abs(self.d) > 1e-13):
            
            w_a = self.scale * (self.bc + cross(p, self.bmc))
            w_b = self.scale * (self.ca + cross(p, self.cma))
            w_c = self.scale * (self.ab + cross(p, self.amb))
            
            is_in = (0 <= w_a <= 1 and 0 <= w_b <= 1 and 0 <= w_c <= 1 )
            
            return (is_in, w_a, w_b, w_c)
        
        else:
            return (False, 0,0,0)
        
        
class MultiTriangleTester:
    
    def __init__(self, a, b, c, width, height):
        
        ''' A class for testing if many points belong to the interior of a triangle
            in two dimensions 
            
            (params):
                a, b, c (array-like): The three points in 2D that 
                        describe the triangle
        '''
        
        # By precalculating everything not involving the test point we can
        # speed up the code significantly
        self.ab = cross(a,b)
        self.bc = cross(b,c)
        self.ca = cross(c,a)
        
        self.d = max(1e-14, self.ab + self.bc + self.ca) # To avoid div by 0
        self.scale = 1/self.d
        
        self.amb = a-b
        self.bmc = b-c
        self.cma = c-a
        
        self.width = width
        self.height = height
        
        # Vectorise everthing and let numpy do the heavy lifting
        self.AMB = np.zeros((height, width, 3))
        self.BMC = np.zeros((height, width, 3))
        self.CMA = np.zeros((height, width, 3))
        for u in range(width):
            for v in range(height):
                self.AMB[v,u,0:2] = a-b
                self.BMC[v,u,0:2] = b-c
                self.CMA[v,u,0:2] = c-a
        
    def test(self, P):
        
        ''' Tests if the points P are within the interior of the triangle defined
            for this class and gives the barycentric co-ordinates they are
            (params):
                p (ndarray, WxHx2): The points that are to be tested
            (returns):
                is_in (ndarray, WxH): If each point is in the triangle
                w_a, w_b, w_c (ndarray, WxH): The barycentric co-ordinates
        '''
        zeros = np.zeros((self.width, self.height))
        
        if (abs(self.d) > 1e-13):
            
            P = np.concatenate((P, np.zeros((self.height, self.width, 1))), axis=2)
            w_a = self.scale * (self.bc + np.cross(P, self.BMC))[:,:,2]
            w_b = self.scale * (self.ca + np.cross(P, self.CMA))[:,:,2]
            w_c = self.scale * (self.ab + np.cross(P, self.AMB))[:,:,2]
            
            is_in = 0 <= w_a
            is_in &= w_a <= 1
            is_in &= 0 <= w_b
            is_in &= w_b <= 1
            is_in &= 0 <= w_c
            is_in &= w_c <= 1
            #is_in = (0 <= w_a <= 1 and 0 <= w_b <= 1 and 0 <= w_c <= 1 )
            
            return (is_in, w_a, w_b, w_c)
        
        else:
            return (False, zeros,zeros,zeros)
        
        
def get_normals(vertices, faces):
    
    ''' This code calculates vertex normals without smoothing groups (yet) 
        
        (params):
            vertices (ndarray, Nx3): The vertices
            faces (ndarray, Nx3x3): The faces 
        (returns):
            normals (ndarray, Nx3): The normals listed in the same order 
                                        as the vertices
                                        
    '''
    
    normals = np.zeros(vertices.shape)
    f_counts = np.zeros((normals.shape[0])) # Counts for averaging
    for face in faces:
        
        ia,ib,ic = face[:, 0]
        a,b,c = vertices[ia], vertices[ib], vertices[ic]
        na, nb, nc = face[:, 2]
        
        B,C = b-a,c-a
        n = np.cross(B,C)
        normals[na] += n
        f_counts[ia] += 1
        normals[nb] += n
        f_counts[ib] += 1
        normals[nc] += n
        f_counts[ic] += 1
        
    for i in range(len(f_counts)):
        
        normals[i] /= f_counts[i]
        
    for normal in normals:

        normal /= np.linalg.norm(normal)
        
    return normals



def apply_rpy(mesh, roll, pitch, yaw):
    
    ''' Rotates, in-place, the mesh by given roll, pitch yaw (in radians)'''
    
    R_z = np.array([[cos(yaw), -sin(yaw), 0],
                    [sin(yaw), cos(yaw), 0 ],
                    [0,          0,       1]
        ])
    
    R_y = np.array([[cos(pitch), 0, sin(pitch)],
                    [0       , 1,        0],
                    [-sin(pitch), 0, cos(pitch)]
        ])
    
    R_x= np.array([[1, 0, 0],
                  [0, cos(roll), -sin(roll)],
                  [0, sin(roll), cos(roll)]
        ])
        
    mesh.apply_transformation(R_x)
    mesh.apply_transformation(R_y)
    mesh.apply_transformation(R_z)
      


import numpy as np



def read_wrap_landmark(f_name):
    
    """ Reads a landmark file from wrap """
    
    # Get the data
    with open(f_name, 'r') as f:
        data = f.read()
    
    # Strip all irrelevant chars
    d_copy = ''
    for c in data:
        if c not in '[ ]':
            d_copy += c
    
    # Split into just numbers
    data = d_copy.split(',')
    
    # Convert into landmarks
    landmarks = []
    for i in range(len(data)//3):
        landmark = []
        landmark.append(int(data[3*i]))     # Index
        landmark.append(float(data[3*i+1])) # a
        landmark.append(float(data[3*i+2])) # b
        landmarks.append(landmark)
        
    return landmarks

def read_wrap_polys_to_vertex_mask(f_name, mesh):
    
    """ Convert a wrap polygon file to a vertex mask """
    
    
    mask = set([])
    face_pos = mesh.faces[:,:,0]
    with open(f_name, 'r') as f:
        for count, line in enumerate(f):
            line = line.strip()
            line = line.split(',')
            d = line[0]
            try:
                x = int(d)
                face = face_pos[x]
                a,b,c = face
                mask.add(a)
                mask.add(b)
                mask.add(c)
            finally:
                continue
            
    return list(mask)


from PyFaceGen.Mesh.Mesh import Mesh

class MeshGraph:
    
    def __init__(self, mesh):
        
        """ A graph represntation of a mesh """
        
        self.vertices = mesh.vertices
        self.edges = {i:set({}) for i in range(len(mesh.vertices))}
        self.n_vertices = self.vertices.shape[0]
        self.n_edges = 0
        
        for face in mesh.faces[:,:,0]:
            self.edges[face[0]].add(face[1])
            self.edges[face[0]].add(face[2])
            
            self.edges[face[1]].add(face[0])
            self.edges[face[1]].add(face[2])
            
            self.edges[face[2]].add(face[1])
            self.edges[face[2]].add(face[0])
            
            self.n_edges += 6
            
    def print(self):
        
        for key in self.edges:
            
            line = "Index %s, value %s: {"%(key, self.vertices[key])
            for edge in self.edges[key]:
                line += str(edge)+" "
                
            line+= "}"
            print(line)

    
    



# python version 3.8
# This code converts between .ply files and our Mesh class
# Large parts of this code are copied from
# https://github.com/daavoo/pyntcloud/blob/master/pyntcloud/io/ply.py
# Available under The MIT License

# PLY LOADING ONLY SUPPORTS POSITIONS, NOT UV OR NORMALS

import sys
import numpy as np
import pandas as pd
from collections import defaultdict
import plyfile
from PyFaceGen.Mesh.Mesh import Mesh
from PyFaceGen.Mesh.Utils import get_normals
import time

sys_byteorder = ('>', '<')[sys.byteorder == 'little']

ply_dtypes = dict([
    (b'int8', 'i1'),
    (b'char', 'i1'),
    (b'uint8', 'u1'),
    (b'uchar', 'b1'),
    (b'uchar', 'u1'),
    (b'int16', 'i2'),
    (b'short', 'i2'),
    (b'uint16', 'u2'),
    (b'ushort', 'u2'),
    (b'int32', 'i4'),
    (b'int', 'i4'),
    (b'uint32', 'u4'),
    (b'uint', 'u4'),
    (b'float32', 'f4'),
    (b'float', 'f4'),
    (b'float64', 'f8'),
    (b'double', 'f8')
])

valid_formats = {'ascii': '', 'binary_big_endian': '>',
                 'binary_little_endian': '<'}

# From pyntcloud
def read_ply(filename):
    """ Read a .ply (binary or ascii) file and store the elements in pandas DataFrame
    Parameters
    ----------
    filename: str
        Path to the filename
    Returns
    -------
    data: dict
        Elements as pandas DataFrames; comments and ob_info as list of string
    """

    with open(filename, 'rb') as ply:

        if b'ply' not in ply.readline():
            raise ValueError('The file does not start whith the word ply')
        # get binary_little/big or ascii
        fmt = ply.readline().split()[1].decode()
        # get extension for building the numpy dtypes
        ext = valid_formats[fmt]

        line = []
        dtypes = defaultdict(list)
        count = 2
        points_size = None
        mesh_size = None
        has_texture = False
        while b'end_header' not in line and line != b'':
            line = ply.readline()

            if b'element' in line:
                line = line.split()
                name = line[1].decode()
                size = int(line[2])
                if name == "vertex":
                    points_size = size
                elif name == "face":
                    mesh_size = size

            elif b'property' in line:
                line = line.split()
                # element mesh
                if b'list' in line:

                    if b"vertex_indices" in line[-1] or b"vertex_index" in line[-1]:
                        mesh_names = ["n_points", "v1", "v2", "v3"]
                    else:
                        has_texture = True
                        mesh_names = ["n_coords"] + ["v1_u", "v1_v", "v2_u", "v2_v", "v3_u", "v3_v"]

                    if fmt == "ascii":
                        # the first number has different dtype than the list
                        dtypes[name].append(
                            (mesh_names[0], ply_dtypes[line[2]]))
                        # rest of the numbers have the same dtype
                        dt = ply_dtypes[line[3]]
                    else:
                        # the first number has different dtype than the list
                        dtypes[name].append(
                            (mesh_names[0], ext + ply_dtypes[line[2]]))
                        # rest of the numbers have the same dtype
                        dt = ext + ply_dtypes[line[3]]

                    for j in range(1, len(mesh_names)):
                        dtypes[name].append((mesh_names[j], dt))
                else:
                    if fmt == "ascii":
                        dtypes[name].append(
                            (line[2].decode(), ply_dtypes[line[1]]))
                    else:
                        dtypes[name].append(
                            (line[2].decode(), ext + ply_dtypes[line[1]]))
            count += 1

        # for bin
        end_header = ply.tell()

    data = {}

    if fmt == 'ascii':
        top = count
        bottom = 0 if mesh_size is None else mesh_size

        names = [x[0] for x in dtypes["vertex"]]

        data["points"] = pd.read_csv(filename, sep=" ", header=None, engine="python",
                                     skiprows=top, skipfooter=bottom, usecols=names, names=names)

        for n, col in enumerate(data["points"].columns):
            data["points"][col] = data["points"][col].astype(
                dtypes["vertex"][n][1])

        if mesh_size :
            top = count + points_size

            names = np.array([x[0] for x in dtypes["face"]])
            usecols = [1, 2, 3, 5, 6, 7, 8, 9, 10] if has_texture else [1, 2, 3]
            names = names[usecols]

            data["mesh"] = pd.read_csv(
                filename, sep=" ", header=None, engine="python", skiprows=top, usecols=usecols, names=names)

            for n, col in enumerate(data["mesh"].columns):
                data["mesh"][col] = data["mesh"][col].astype(
                    dtypes["face"][n + 1][1])

    else:
        with open(filename, 'rb') as ply:
            ply.seek(end_header)
            points_np = np.fromfile(ply, dtype=dtypes["vertex"], count=points_size)
            if ext != sys_byteorder:
                points_np = points_np.byteswap().newbyteorder()
            data["points"] = pd.DataFrame(points_np)
            if mesh_size:
                mesh_np = np.fromfile(ply, dtype=dtypes["face"], count=mesh_size)
                if ext != sys_byteorder:
                    mesh_np = mesh_np.byteswap().newbyteorder()
                data["mesh"] = pd.DataFrame(mesh_np)
                data["mesh"].drop('n_points', axis=1, inplace=True)

    return data


def write_ply(filename, points=None, mesh=None, as_text=False):
    
    """
    Parameters
    ----------
    filename: str
        The created file will be named with this
    points: ndarray
    mesh: ndarray
    as_text: boolean
        Set the write mode of the file. Default: binary
    Returns
    -------
    boolean
        True if no problems
    """
    
    points = pd.DataFrame({'x': points[:,0], 'y':points[:,1], 'z':points[:,2]})
    mesh = pd.DataFrame(mesh)
    
    if not filename.endswith('ply'):
        filename += '.ply'

    # open in text mode to write the header
    with open(filename, 'w') as ply:
        header = ['ply']

        if as_text:
            header.append('format ascii 1.0')
        else:
            header.append('format binary_' + sys.byteorder + '_endian 1.0')

        if points is not None:
            header.extend(describe_element('vertex', points))
        if mesh is not None:
            mesh = mesh.copy()
            mesh.insert(loc=0, column="n_points", value=3)
            mesh["n_points"] = mesh["n_points"].astype("u1")
            header.extend(describe_element('face', mesh))

        header.append('end_header')

        for line in header:
            ply.write("%s\n" % line)

    if as_text:
        if points is not None:
            points.to_csv(filename, sep=" ", index=False, header=False, mode='a',
                          encoding='ascii')
        if mesh is not None:
            mesh.to_csv(filename, sep=" ", index=False, header=False, mode='a',
                        encoding='ascii')

    else:
        with open(filename, 'ab') as ply:
            if points is not None:
                points.to_records(index=False).tofile(ply)
            if mesh is not None:
                mesh.to_records(index=False).tofile(ply)

    return True

def describe_element(name, df):
    
    """ Takes the columns of the dataframe and builds a ply-like description
    Parameters
    ----------
    name: str
    df: pandas DataFrame
    Returns
    -------
    element: list[str]
    """
    property_formats = {'f': 'float', 'u': 'uchar', 'i': 'int'}
    element = ['element ' + name + ' ' + str(len(df))]

    if name == 'face':
        element.append("property list uchar int vertex_indices")

    else:
        for i in range(len(df.columns)):
            # get first letter of dtype to infer format
            f = property_formats[str(df.dtypes[i])[0]]
            element.append('property ' + f + ' ' + df.columns.values[i])

    return element

def ply_2_mesh(path, compute_normals = True):
    
    # Read to pandas
    data = read_ply(path)
    
    vertices = data['points'].to_numpy()

    if vertices.shape[-1] >= 3:
        # Properties other than position
        vertices = vertices[:,0:3]
        
    UVs = np.zeros((vertices.shape[0], 2)) 
    faces = data['mesh'].to_numpy().reshape((-1, 3, 1))
    if compute_normals:
        normals = get_normals(vertices, faces)
    else:
        normals = np.zeros(vertices.shape)
  
    mesh = Mesh(vertices, UVs, normals, faces)
    return mesh

def mesh_2_ply(mesh, path):
    
    write_ply(path, mesh.vertices, mesh.faces, as_text=True)





import pyassimp
import numpy as np
from PyFaceGen.Mesh.Mesh import Mesh

def pyassimp_2_mesh(path):
    
    ''' Converts a .obj file to the custom mesh class '''
    
    
    # Load the scene with pyassimp
    scene = pyassimp.load(path,  
                  processing=pyassimp.postprocess.aiProcess_JoinIdenticalVertices)
    
    # Extract the mesh and convert to our class
    mesh = scene.meshes[0]
    
    vertices = mesh.vertices
        
    if len(mesh.texturecoords) == 0:
        UVs = np.zeros((mesh.vertices.shape[0], 2))
    else:
        UVs = mesh.texturecoords[0,:,0:2]
    normals = mesh.normals
    faces = mesh.faces
    
    mesh = Mesh(vertices, UVs, normals, faces)
    
    return mesh


"""  

THIS CODE SHOULD NOT BE NEEDED
      
def force_pyassimp_uvs(path):
    
    mesh_no_uv = pyassimp_2_mesh(path)
    mesh_uvs_not_joined = obj_2_mesh(path)

    j = 0 # Index of vertex in non-joined mesh
    for i in range(mesh_no_uv.vertices.shape[0]):
        
        v1 = mesh_no_uv.vertices[i]
        v2 = mesh_uvs_not_joined.vertices[j]
        while np.linalg.norm(v1-v2) > 0.01:
            j += 1 
            v2 = mesh_uvs_not_joined.vertices[j]
        mesh_no_uv.UVs[i] = mesh_uvs_not_joined.UVs[j]
        j += 1
        
    return mesh_no_uv
"""
    



from PyFaceGen.Mesh.Mesh import Mesh
from PyFaceGen.Mesh.Graph import MeshGraph
from PyFaceGen.IO.wobj_loader import wobj_2_mesh

def GraphTest():
    
    cube = wobj_2_mesh('Data/Tests/primatives/cube.obj')
    graph = MeshGraph(cube)
    assert graph.edges[0] == set((1,2,4,5,6))



# python version 3.8
# This code describes the mesh class which we will work with

import numpy as np

class Mesh:
    
    def __init__(self, vertices, UVs, normals, faces, dtype='float32'):
        
        ''' The custom Mesh class that we will work with
        
            (params):
                verticies (ndarray): The positions of the vertices in 3D
                UVs (ndarray): The UV co-ordinates of each vertex
                normals (ndarray): The normals to each vertex
                faces (ndarray nFacesx3x3): The indexes of the vertices making up each face
                    faces[:,:,0] is positions, faces[:,:,1] is UVs, faces[:,:,2] is normals
                
        '''
        
        self.vertices = np.array(vertices, dtype=dtype)
        self.UVs = np.array( UVs, dtype=dtype)
        self.normals = np.array(normals, dtype=dtype)
        self.faces = faces
        
        self.n_vertices = len(self.vertices)
        
        self.order = []
        
    def update_buffer(self):
        
        ''' Updates the internal buffer that is used GLFW '''
        
        buffer = []
        for idx in self.faces.shape[0]:
            
            face = self.faces[idx,:,:] # 3x3
            
            buffer.extend(self.vertices[face[0,0]])
            buffer.extend(self.UVs[face[0,1]])
            buffer.extend(self.normals[face[0,2]])
            
            buffer.extend(self.vertices[face[1,0]])
            buffer.extend(self.UVs[face[1,1]])
            buffer.extend(self.normals[face[1,2]])
            
            buffer.extend(self.vertices[face[1,0]])
            buffer.extend(self.UVs[face[1,1]])
            buffer.extend(self.normals[face[2,2]])
            
        self.buffer = np.array(buffer, dtype='float32')
        
    def apply_translation(self, vector):
        
        ''' Applies a translation of the mesh by a given vector 
        
            (params): 
                vector (3x1 numpy array): The 3D translation vector 
        '''
        
        # Note that only vertex positions are altered, we do not change the UV
        # or the normals
        self.vertices += vector
        
        
    def apply_transformation(self, matrix):
        
        ''' Applies a linear transformation of the mesh defined by
            a given 3x3 matrix 
        
            (params): 
                matrix (3x3 numpy array): The 3D transformation matrix 
        '''
        
        # Here the normals need to be updated as well as the positions
        self.vertices = self.vertices @ matrix
        norm_transform = np.linalg.inv(matrix.T)        
        self.normals = self.normals @ norm_transform
        
    
        


# python version 3.8
# This code is used to align any two arbitrary meshes using rotation, translation
# and scale. 

# We use ICP with additional scaling as done inhttp://web.stanford.edu/class/cs273/refs/umeyama.pdf 

import numpy as np
from sklearn.neighbors import NearestNeighbors
from PyFaceGen.Mesh.Mesh import Mesh
from PyFaceGen.Mesh.Utils import get_normals

def estimateRigidTransform(A,B):
    
    """ Estimates the optimal 4x4 rigid transform between points in correspondence
    
        (returns):
            T (ndarray of 4x4): The transformation matrix 
    
    """
    
    assert A.shape == B.shape, "A and B must be the same shape but are %s and %s"%(A.shape, B.shape)
    
    n = A.shape[0]
    
    # Compute centroids and shift the points
    centroid_A = np.mean(A, axis=0)
    centroid_B = np.mean(B, axis=0)
    
    A_ = A - centroid_A
    B_ = B - centroid_B
    
    # Compute the H matrix and perform single value decomposition
    H = A_.T @ B_ / n
    U, D, VT = np.linalg.svd(H)
    
    # Compute the optimal rotation which is VU^T
    R = VT.T @ U.T
    # special reflection case
    if np.linalg.det(R) < 0:
       VT[2,:] *= -1
       D[-1] *= 1
       R = VT.T @ U.T
    
    varP = np.var(A, axis=0).sum()
    scale = 1/varP * np.sum(D)
    
    '''if np.isnan(scale):
        T = np.eye(4)
        return T, np.eye(3), 1, np.zeros((3,1))'''
    
    RS = R * scale
    #print(scale)
    # Compute translation
    t = centroid_B.T - (RS @ centroid_A.T)
    
    
    T = np.eye(4)
    T[0:3,0:3] = RS
    T[0:3, 3] = t
    return T, R, scale, t
    
    
    

def rigid_align(fixed, floating, n_iters=10000, init_rotation=np.eye(3), init_translation=np.zeros((3,)), init_scale=np.ones((3,)), tolerance=0.001):
    
    """ Rigidly align, using rotation, translation and scale, a floating mesh
        to a fixed mesh 
        
        (params):
            fixed (Mesh): The mesh that will not be moving
            floating (Mesh): The mesh that will be transformed            
        (optional params)
            n_iters (int): The number of iterations
            init_rotation (3x3 ndarray): Provides an inital estimate for rotataion as a speed up
            init_translation (3, ndarray): Provides an inital estimate for translation as a speed up
            init_scale (3, ndarray): Provides an inital estimate for scale as a speed up
            tolerance (float): The tolerance at which to end the algorithm early
        (returns):
            (Mesh): The floating mesh, aligned to the fixed mesh
            (2d-ndarray of 4x4): The transformation matrix applied to the floating mesh
    
    """
    
    # Get the vertices and transpose
    
    f_copy = floating.vertices

    source = np.float32(floating.vertices.T)
    target = np.float32(fixed.vertices.T)
    
    # Convert to homogenous co-ordinates
    source = np.concatenate((source, np.ones((1, source.shape[1]), dtype='float32' )))
    target = np.concatenate((target, np.ones((1, target.shape[1]), dtype='float32' )))

    # Compute the 4x4 transformation matrix for the inital transformation    
    transform = np.eye(4)
    transform[0:3,0:3] = init_rotation * init_scale
    transform[0:3,3] = init_translation
    
    # Transform the source by the initial estimate
    source = transform @ source
    
    # Keep track of how good an estimate we have
    error = np.inf
    best_error = error
    best_transform = np.array(transform)
    
    for i in range(n_iters):
        # First we compute point correspondences
        
        nn = NearestNeighbors(n_neighbors=1).fit(target.T)
        #print(np.max(source.T))
        
        dist, idx = nn.kneighbors(source.T)
        
        # Get the optimal rigid transform for these point correspondences
        transform, _,s,_ = estimateRigidTransform(source[0:3,:].T, target[0:3, idx].reshape(3,-1).T)
        #print(s)
        
        # Do the transform
        s_copy = np.array(source)
        
        source = transform @ source
        
        new_error = np.sum(dist)/source.shape[0]
        
        if abs(error - new_error) < tolerance:
            break
        error = new_error
    
    verts = source.T[:,0:3]
    
    
    floating.vertices = verts
    
    # Transform the normals as well
    #norms = np.concatenate([floating.normals, np.ones((floating.normals.shape[0], 1))], axis=1).T
    #norms = np.linalg.inv(transform).T @ norms
    
    floating.normals = get_normals(floating.vertices, floating.faces)
    
    return floating

def rigid_align_with_points(fixed, floating, fi_landmarks, fl_landmarks, n_iters=100):
    
    """ Rigidly align the floating mesh to the fixed mesh using landmarks 
        in correspondance.
        
        A landmark should look like [id, a, b] where id is the index of
            the triangle containing the landmark and a,b its barycentric coords
    """
    
    if len(fl_landmarks) != len(fi_landmarks):
        
        raise RuntimeError('Must have a corresponance between the landmarks')
    
    if len(fl_landmarks) < 3:
        
        raise RuntimeError('Must have at least 3 landmarks')
        
    # First we get the 3D points from the landmarks using the barycentric coordinates
    fl_points = []
    fi_points = []
    
    for i in range(len(fl_landmarks)):
        fl_land, fi_land = fl_landmarks[i], fi_landmarks[i]
        
        fl_id, fl_a, fl_b = fl_land
        fi_id, fi_a, fi_b = fi_land
        
        # First the floating landmark
        a,b,c = floating.faces[fl_id,:,0]
        va,vb,vc = floating.vertices[a], floating.vertices[b], floating.vertices[c]
        
        fl_point = (fl_a * va) + (fl_b *vb) +((1- fl_a - fl_b) *  vc)
        fl_points.append(fl_point)
        
        # Then the fixed
        a,b,c = fixed.faces[fi_id,:,0]
        va,vb,vc = fixed.vertices[a], fixed.vertices[b], fixed.vertices[c]
        
        fi_point = (fi_a * va) + (fi_b *vb) +((1- fi_a - fi_b) *  vc)
        fi_points.append(fi_point)
        
    fl_points = np.array(fl_points)
    fi_points = np.array(fi_points)
    
    _, R, s, t = estimateRigidTransform(fl_points, fi_points)
    
    #print(R, s, t)
    
    return rigid_align(fixed, floating, n_iters, R, t, s)



import numpy as np

from PyFaceGen.Mesh.Mesh import Mesh
from PyFaceGen.Mesh.Graph import MeshGraph
from PyFaceGen.IO.wobj_loader import wobj_2_mesh, mesh_2_wobj
from PyFaceGen.Alignment.RigidAlignment import rigid_align, rigid_align_with_points
from PyFaceGen.Alignment.landmarks import read_wrap_landmark
from PyFaceGen.Alignment.NRICP import NRICP, Transform, get_d_loss, get_s_loss

def rigid_test():
    
    fixed = wobj_2_mesh('Data/Faces/rig_neutral.obj')
    floating = wobj_2_mesh('Data/Faces/lyhm_1.obj')
    
    fl_land = read_wrap_landmark('Data/Faces/flame_landmarks.txt')
    fi_land = read_wrap_landmark('Data/Faces/rig_landmarks.txt')
    
    mesh = rigid_align_with_points(fixed, floating, fi_land, fl_land)
    mesh_2_wobj(mesh, "Data/Faces/align.obj")

def NRICP_transform_test():
    
    # Load a cube in homogenous co-ordinates
    cube = wobj_2_mesh('Data/Tests/primatives/cube.obj') 
    V = np.concatenate((cube.vertices, np.ones((cube.vertices.shape[0], 1))), axis=1)
    transform = Transform(cube.vertices.shape[0])
    V_ = transform.apply_transform(V)

    assert np.sum(abs(V - V_)) == 0    

def NRICP_d_loss_test():
    
    # Load a cube in homogenous co-ordinates
    cube = wobj_2_mesh('Data/Tests/primatives/cube.obj') 
    V = np.concatenate((cube.vertices, np.ones((cube.vertices.shape[0], 1))), axis=1)
    transform = Transform(cube.vertices.shape[0])
    V_ = transform.apply_transform(V)

    assert get_d_loss(V,V_) == 0
    
    V_[0,0] = V_[0,0]+1 
    assert get_d_loss(V, V_) == 1
    
def NRICP_g_loss_test():
    
    # Load a cube in homogenous co-ordinates
    cube = wobj_2_mesh('Data/Tests/primatives/cube.obj') 
    V = np.concatenate((cube.vertices, np.ones((cube.vertices.shape[0], 1))), axis=1)
    transform = Transform(cube.vertices.shape[0])
    V_ = transform.apply_transform(V)
    graph = MeshGraph(cube)
    assert get_s_loss(graph, transform, 1) == 0
    
def NRICP_test():
    
    cube = wobj_2_mesh('Data/Tests/primatives/cube.obj') 
    sphere = wobj_2_mesh('Data/Tests/primatives/sphere.obj')
    cube_landmarks = read_wrap_landmark('Data/Tests/primatives/cube_landmarks.txt')
    sphere_landmarks = read_wrap_landmark('Data/Tests/primatives/sphere_landmarks.txt')
    sphere_cube = NRICP(sphere, cube, sphere_landmarks, cube_landmarks, verbose=True)
    mesh_2_wobj(sphere_cube , 'Data/Tests/primatives/sphere_cube.obj')
    
def non_rigid_test():
    
    NRICP_transform_test()
    NRICP_d_loss_test()
    NRICP_g_loss_test()
    NRICP_test()



# Imports

import numpy as np
import matplotlib.pyplot as plt
import PIL.Image as Image
import time
from sklearn.neighbors import KNeighborsRegressor


from PyFaceGen.Mesh.Mesh import Mesh
from PyFaceGen.Mesh.Utils import TriangleTester, MultiTriangleTester
from PyFaceGen.IO.wobj_loader import wobj_2_mesh, mesh_2_wobj

class AlbedoPCA:
    
    def __init__(self, path):
        
        self.model = np.load(path)
        
        self.mu = self.model['MU']
        
        b,g,r = self.mu[:,:,0], self.mu[:,:,1], self.mu[:,:,2]
        self.mu = np.stack((r,g,b), axis=2)
        
        self.PC = self.model['PC']
        b,g,r = self.PC[:,:,0,:], self.PC[:,:,1,:], self.PC[:,:,2,:] 
        self.PC[:,:,0,:] = r
        self.PC[:,:,1,:] = g
        self.PC[:,:,2,:] = b
        
        self.s_mu = self.model['specMU']
        
        b,g,r = self.s_mu[:,:,0], self.s_mu[:,:,1], self.s_mu[:,:,2]
        self.s_mu = np.stack((r,g,b), axis=2)
        
        self.s_PC = self.model['specPC']
        b,g,r = self.s_PC[:,:,0,:], self.s_PC[:,:,1,:], self.s_PC[:,:,2,:] 
        self.s_PC[:,:,0,:] = r
        self.s_PC[:,:,1,:] = g
        self.s_PC[:,:,2,:] = b
        
        
        
    def get_diffuse_tex(self, weights):
        
        return self.mu + self.PC.dot(weights)
    
    def get_diffuse_tex_one_weight(self, idx, weight):
        
        return self.mu + weight*self.PC[:,:,:,idx]
    
    def get_spec_tex(self, weights):
        
        return self.s_mu + self.s_PC.dot(weights)
    
    def get_spec_tex_one_weight(self, idx, weight):
        
        return self.s_mu + weight*self.s_PC[:,:,:,idx]

class Excluder:
    
    """ This class determines if a texel should be excluded from the projections 
        using lines 
    """
    
    def __init__(self):
        
        self.lines=[]
        
    def add_line(self, slope, intercept, exclude_side):
        
        """ Adds a line used to exclude texels 
            (params):
                slope (float): The slope of the line
                intercept (float): The y intercept of the line
                exclude_side (+-1): Which side of the line to exclude. +1 means above, -1 below
        """
        
        self.lines.append([slope, intercept, exclude_side])
        
    def add_line_points(self, u1,v1,u2,v2, exclude_side):
        
        slope = (u2-u1)/(v2-v1)
        intercept = u2 - slope*v2
        
        print(slope, intercept)
        
        self.add_line(slope, intercept, exclude_side)
        
    def should_exclude(self, u,v):
        
        for line in self.lines:
            # Get y(x) for the given line
            Y = line[0]*u + line[1]
            diff = v-Y # Positive if point is above the line
            diff *= line[2] # Now positive if we want to exclude the point
            
            if diff>0:
                return True
        return False    
        
    def test(self, image):
        
        """ Shows an example of which texels are excluded from a given texture"""
        
        for v in range(image.height):
            for u in range(image.width):
                if self.should_exclude(u,v):
                    image.putpixel((v,u), (0,0,255))
                    
        plt.imshow(image)
        plt.show()

def project_texture_to_3d(mesh, texture_map, excluder=None):
    
    """ This code projects the UV space into 3 dimensions and provides an
        estimate correspondance between 3D pos and color
        (params): 
            mesh (Mesh): The mesh for which we project UVs
            texture_map: The texture
    """
    
    # Any uv not inside a face can just not be projected
    
    projections = []
    colours = []
    pix_coords = []
    
    
    # Get and scale the UVs to the image
    UVs = np.array(mesh.UVs)
    UVs[:,0] *= texture_map.width - 1
    UVs[:,1] *= texture_map.height - 1
    
    for i,face in enumerate(mesh.faces):
        
        # Iterate over the faces
        a,b,c = face
        uva, uvb, uvc = UVs[a], UVs[b], UVs[c]
        pa, pb, pc = mesh.vertices[a], mesh.vertices[b], mesh.vertices[c]
        
        # Get the bounding box of this triangle in UV space
        min_u = int(round(min(uva[0], uvb[0], uvc[0])))
        max_u = int(round(max(uva[0], uvb[0], uvc[0])))
        min_v = int(round(min(uva[1], uvb[1], uvc[1])))
        max_v = int(round(max(uva[1], uvb[1], uvc[1])))
        
        
        # To catch the triangles the are go across patches in UV space we set a limit on triangle size
        
        if max_u - min_u > 0.2*texture_map.width or max_v - min_v > 0.2*texture_map.height:
            continue
        
        # The vertices of the triangle in uv space
        x = np.round(uva)
        y = np.round(uvb)
        z = np.round(uvc)
        
        width = max_u - min_u + 1
        height = max_v - min_v + 1
        
        # Initialise the triangle tester for this face
        tester = MultiTriangleTester(x,y,z,width,height)
        
        # Get all points for testing
        P = np.zeros((height, width, 2))
        for j,u in enumerate(range(min_u, max_u+1)):
            for k,v in enumerate(range(min_v, max_v+1)):
                P[k,j] = np.array([u,v])
                
        is_in, w_a, w_b, w_c = tester.test(P)
        if np.isscalar(is_in):
            # Prevents runtime error when only 1 pixel is considered
            continue
            
        for j, u in enumerate(range(min_u, max_u+1)):  # +1 so inclusive
            for k, v in enumerate(range(min_v, max_v+1)): 
                
                if excluder is not None:
                    if excluder.should_exclude(u, texture_map.height-v):
                        continue
                    
                try:
                    if is_in[k,j]:
                        
                        position = w_a[k,j] * pa + w_b[k,j] * pb + w_c[k,j] * pc
                        color = texture_map.getpixel((u,texture_map.height-v))
                        
                        projections.append(position)
                        colours.append(color)
                        pix_coords.append(np.array([u,texture_map.height-v]))
                        
                except:
                    print(is_in)
                    raise
                    
        if i % 1000 ==0:
            
            print('%s out of %s'%(i, mesh.faces.shape[0]))
    return np.array(projections), np.array(colours), np.array(pix_coords)



def transfer_texture(source_mesh, target_mesh, source_texture, target_texture_size, plot=True):
    
    """ Transfers the source texture to the target texture using KNN given 2 meshes as 
        closely aligned as possible
        (params):
            source_mesh (Mesh): The mesh corresponding to the known texture
            target_mesh (Mesh): The mesh corresponding to the desired texture
            source_texture (PIL Image): The known texture
            target_texture_size (2-Tuple): The size of the output target texture (width, height)
    """
    
    # Create a blank
    target_texture = Image.new('RGB', target_texture_size)
    
    # First project each texel in the target texture to 3D space
    t_pos, t_tex, t_coords = project_texture_to_3d(target_mesh, target_texture) 
    
    # Then project the source texels
    s_pos, s_tex, s_coords = project_texture_to_3d(source_mesh, source_texture)
    
    # Now fit a KNN regressor to the source mesh to produce a color map in 3D space
    model = KNeighborsRegressor()
    model.fit(s_pos, s_tex)
    
    # Use this map to estimate the target texels from their 3D projection
    t_tex = model.predict(t_pos)
    
    for i in range(t_coords.shape[0]):
        # Set the texel to its predicted value in the texture
        coord = t_coords[i]
        u,v = coord
        colour = t_tex[i]
        r,g,b = colour
        r = int(r)
        g = int(g)
        b = int(b)
        target_texture.putpixel((u,v),(r,g,b))
   
    if plot:
        plt.imshow(target_texture)
        plt.show()
    
    return target_texture

def main(plot=True):
    
    """ Specifically for AlbedoMM --> rig """
    
    # Load textures
    rig_texture = Image.open('C:\\Users\\jacks\\OneDrive\\Documents\\MRES\\Project\\Summer - Corrective Blendshapes\\ObjRenderer\\data\\male_diffuse.png')
    mm_texture = Image.open('D:\\Data\\FLAME_Topology\\Albedo\\FLAME_DIFFUSE.png')

    # Load objects
    rig_mesh = wobj_2_mesh('C:/Users/jacks/OneDrive/Documents/PhD/Expression Wrapping/blendshapes/transferred_with_uvs/blend_0.obj')
    lyhm_mesh = wobj_2_mesh('C:/Users/jacks/OneDrive/Documents/PhD/Expression Wrapping/rigid_target.obj')
    
    # Define the texels we do not want to project
    exclude = Excluder()
    exclude.add_line_points(2600,4096, 4096, 1300,1)
    exclude.add_line_points(0,740, 670, 0,-1)
    exclude.add_line_points(4096,770, 4096-660, 0,1)
    
    # ONLY USE THIS LINE AS A TEST, DO NOT RUN WITH THE REST
    #exclude.test(rig_texture)
    
    # Load the Albedo MM for transfer
    # https://github.com/waps101/AlbedoMM
    mm = AlbedoPCA('D:/Data/FLAME_Topology/Albedo/albedoModel2020_FLAME_albedoPart.npz')
    
    # PIL Image format
    mm_texture = Image.fromarray(np.uint8(255*mm.mu))
    plt.imshow(mm_texture)
    plt.show()
    
    # Project every texel from the target image into 3D space
    print('Projecting target texture')
    start = time.time()
    t_pos, t_tex, t_coords = project_texture_to_3d(rig_mesh, rig_texture, exclude) 
    print('Took %s s'%(time.time() - start))  
    
    # Project every texel from the source image into 3D space
    print('Projecting source texture')
    start = time.time()
    s_pos, s_tex, s_coords = project_texture_to_3d(lyhm_mesh, mm_texture)
    print('Took %s s'%(time.time() - start))
    
    # Fit a KNN regressor to predict albedo from 3D position, trained on the source image
    print('Fitting Model')
    start = time.time()
    model = KNeighborsRegressor()
    model.fit(s_pos, s_tex)
    print('Took %s s'%(time.time() - start))
    
    # Predict the target texels albedo value from its projection in 3D space using the KNN regeressor
    t_tex = model.predict(t_pos)
    
    for i in range(t_coords.shape[0]):
        # Set the texel to its predicted value in the texture
        coord = t_coords[i]
        u,v = coord
        colour = t_tex[i]
        r,g,b = colour
        r = int(r)
        g = int(g)
        b = int(b)
        rig_texture.putpixel((u,v),(r,g,b))
   
    if plot:
        plt.imshow(rig_texture)
        plt.show()
        
    # TODO: Where should we put the output textures
    mean = np.array(rig_texture.getdata(), dtype='uint8').reshape((rig_texture.height, rig_texture.width, 4))[:,:,0:3]
    np.save('textures/mean.npy', mean)
    
    # Now we want to transfer the principal components of the model
    PC = np.zeros((rig_texture.height, rig_texture.width,3,mm.PC.shape[-1]), dtype='uint8')
    for i in range(mm.PC.shape[-1]):
        
        print(np.max(mm.PC[:,:,:,i]), np.min(mm.PC[:,:,:,i]))
        
        rig_texture = Image.open('C:\\Users\\jacks\\OneDrive\\Documents\\MRES\\Project\\Summer - Corrective Blendshapes\\ObjRenderer\\data\\male_diffuse.png')
        mm_texture = Image.fromarray(np.uint8(255*(mm.mu+mm.PC[:,:,:,i])))
        plt.imshow(mm_texture)
        plt.show()
        print('Projecting source texture')
        
        start = time.time()
        s_pos, s_tex, s_coords = project_texture_to_3d(lyhm_mesh, mm_texture)
        
        print('Took %s s'%(time.time() - start))
        
        
        print('Fitting Model')
        
        start = time.time()
        model = KNeighborsRegressor()
        model.fit(s_pos, s_tex)
        
        print('Took %s s'%(time.time() - start))
        
        
            
        t_tex = model.predict(t_pos)
        
        for j in range(t_coords.shape[0]):
            
            coord = t_coords[j]
            u,v = coord
            
            colour = t_tex[j]
            r,g,b = colour
            r = int(r)
            g = int(g)
            b = int(b)
            rig_texture.putpixel((u,v),(r,g,b))

        plt.imshow(rig_texture)
        plt.show()
        
        pc = np.array(rig_texture.getdata()).reshape(rig_texture.height, rig_texture.width, 4)[:,:,0:3]
        plt.imshow(pc-mean)
        plt.show()
        PC[:,:,:,i] = pc-mean
    
    np.save('textures/pcs.npy',PC)
    
     
